#ifndef VM_H_
#define VM_H_

struct vm;

#define	STRING_LEN	256

typedef enum {
	VM_UNKNOWN,
	VM_RUNNING,
	VM_STOPPED,
	VM_EXCEPTION
} vm_state_t;

typedef enum {
	EXCEPT_UNKNOWN = -1,
	EXCEPT_UNKNOWN_OPCODE,
	EXCEPT_UNKNOWN_AREA,
	EXCEPT_OO_STACK,
	EXCEPT_OO_DATA,
	EXCEPT_OO_TEXT,
	EXCEPT_OO_RO,
	EXCEPT_OO_ARRAY,
	EXCEPT_ZERO_DIV,
	EXCEPT_LIBCALL,
	EXCEPT_UKNOWN_REF_POINT,
	EXCEPT_READ_OUTPUTS,
	EXCEPT_WRITE_INPUTS,
	EXCEPT_MEM_AREA_DTYPE,
	EXCEPT_OO_INPUTS,
	EXCEPT_OO_OUTPUTS,
	EXCEPT_BIT_NUM
} except_type_t;

typedef void (*callback_t)(struct vm *machine);
typedef void (*exception_cb_t)(struct vm *machine, except_type_t type, const char *msg);

struct buffer;

struct vm {
	unsigned char	*mmap_begin;
	size_t		mmap_size;

	uint32_t	pc;
	uint32_t	sp;
	uint32_t	fp;
	uint32_t	start_addr;
	uint32_t	stop_addr;
	
	uint32_t	*stack;
	unsigned int	stack_size;
	
	unsigned char	*text;
	unsigned int	text_size;
	
	unsigned char	*init;
	unsigned int	init_size;

	uint32_t 	*data;
	unsigned int	data_size;
	
	unsigned char	*ro;
	unsigned int	ro_size;

	unsigned char	*string;

	struct sym_tb_entry *sym_tb;
	unsigned int	sym_tb_size;

	vm_state_t	state;
	
	struct buffer	*inputs;
	unsigned int	inputs_sz;
	struct buffer	*outputs;
	unsigned int	outputs_sz;
	int		byte_order;

	callback_t	vm_singlestep_callback;
	exception_cb_t	vm_exception_callback;

	void		**dl_handle;
	unsigned int	dl_used;
	unsigned int	dl_size;
};

struct vm *vm_create();
ret_t	vm_destroy(struct vm **machine);

ret_t	vm_prepare(struct vm *machine);
ret_t	vm_load_file(struct vm *machine, const char *file_name,
		     const char *sym_name, int lock_mem);
ret_t	vm_unload_file(struct vm *machine);

ret_t	vm_load_lib(struct vm *machine, const char *path_to_lib);
ret_t	vm_unload_lib(struct vm *machine);
ret_t	vm_io_area_set(struct vm *machine, void *inputs, unsigned int inputs_sz,
		       void *outputs, unsigned int outputs_sz, int byte_order);

ret_t	vm_io_area_unset(struct vm *machine);

void	vm_exec_instr(struct vm *machine);
void	vm_run(struct vm *machine);

void	vm_stack_shrink(struct vm *machine, size_t size);
void	vm_stack_grow(struct vm *machine, size_t size);

int32_t	vm_stack_get_int(struct vm *machine);
float	vm_stack_get_float(struct vm *machine);
char	vm_stack_get_char(struct vm *machine);
void	vm_stack_get_string(struct vm *machine, char str[STRING_LEN]);

int32_t	vm_stack_peek_int(struct vm *machine);
float	vm_stack_peek_float(struct vm *machine);
char	vm_stack_peek_char(struct vm *machine);
void	vm_stack_peek_string(struct vm *machine, char str[STRING_LEN]);

void	vm_stack_put_int(struct vm *machine, int32_t num);
void	vm_stack_put_float(struct vm *machine, float fl);
void	vm_stack_put_char(struct vm *machine, char ch);
void	vm_stack_put_string(struct vm *machine, char str[STRING_LEN]);

void	vm_stack_poke_int(struct vm *machine, int32_t num);
void	vm_stack_poke_float(struct vm *machine, float fl);
void	vm_stack_poke_char(struct vm *machine, char ch);
void	vm_stack_poke_string(struct vm *machine, char str[STRING_LEN]);

#endif /*VM_H_*/
